@@ -17,6 +17,7 @@ except:
     from mercurial.httprepo import passwordmgr
 from mercurial.httprepo import httprepository
 from mercurial import mail
+from urllib2 import AbstractBasicAuthHandler, AbstractDigestAuthHandler
 
 # mercurial.demandimport incompatibility workaround,
 # would cause gnomekeyring, one of the possible
@@ -35,9 +36,13 @@ KEYRING_SERVICE = "Mercurial"
 
 ############################################################
 
-def monkeypatch_method(cls):
+def monkeypatch_method(cls,fname=None):
     def decorator(func):
-        setattr(cls, func.__name__, func)
+        local_fname = fname
+        if local_fname is None:
+            local_fname = func.__name__
+        setattr(func, "orig", getattr(cls, local_fname, None))
+        setattr(cls, local_fname, func)
         return func
     return decorator
 
@@ -77,6 +82,16 @@ class PasswordStore(object):
 
 password_store = PasswordStore()
 
+############################################################
+
+def _debug(ui, msg):
+    ui.debug("[HgKeyring] " + msg + "\n")
+
+def _debug_reply(ui, msg, url, user, pwd):
+    _debug(ui, "%s. Url: %s, user: %s, passwd: %s" % (
+            msg, url, user, pwd and '*' * len(pwd) or 'not set'))
+
+
 ############################################################
 
 class HTTPPasswordHandler(object):
@@ -90,7 +105,7 @@ class HTTPPasswordHandler(object):
         self.pwd_cache = {}
         self.last_reply = None
 
-    def find_auth(self, pwmgr, realm, authuri):
+    def find_auth(self, pwmgr, realm, authuri, req):
         """
         Actual implementation of find_user_password - different
         ways of obtaining the username and password.
@@ -103,7 +118,10 @@ class HTTPPasswordHandler(object):
         # reusing bad password indifinitely).
         after_bad_auth = (self.last_reply \
                           and (self.last_reply['realm'] == realm) \
-                          and (self.last_reply['authuri'] == authuri))
+                          and (self.last_reply['authuri'] == authuri) \
+                          and (self.last_reply['req'] == req))
+        if after_bad_auth:
+            _debug(ui, _("Working after bad authentication, cached passwords not used %s") % str(self.last_reply))
 
         # Strip arguments to get actual remote repository url.
         base_url = self.canonical_url(authuri)
@@ -113,9 +131,9 @@ class HTTPPasswordHandler(object):
         user, pwd = urllib2.HTTPPasswordMgrWithDefaultRealm.find_user_password(
             pwmgr, realm, authuri)
         if user and pwd:
-            self._debug_reply(ui, _("Auth data found in repository URL"),
-                              base_url, user, pwd)
-            self.last_reply = dict(realm=realm,authuri=authuri,user=user)
+            _debug_reply(ui, _("Auth data found in repository URL"),
+                         base_url, user, pwd)
+            self.last_reply = dict(realm=realm,authuri=authuri,user=user,req=req)
             return user, pwd
 
         # Loading .hg/hgrc [auth] section contents. If prefix is given,
@@ -125,7 +143,7 @@ class HTTPPasswordHandler(object):
             keyring_url = prefix_url
         else:
             keyring_url = base_url
-        ui.debug("Keyring URL: %s\n" % keyring_url)
+        _debug(ui, _("Keyring URL: %s") % keyring_url)
 
         # Checking the memory cache (there may be many http calls per command)
         cache_key = (realm, keyring_url)
@@ -133,9 +151,9 @@ class HTTPPasswordHandler(object):
             cached_auth = self.pwd_cache.get(cache_key)
             if cached_auth:
                 user, pwd = cached_auth
-                self._debug_reply(ui, _("Cached auth data found"),
-                                  base_url, user, pwd)
-                self.last_reply = dict(realm=realm,authuri=authuri,user=user)
+                _debug_reply(ui, _("Cached auth data found"),
+                             base_url, user, pwd)
+                self.last_reply = dict(realm=realm,authuri=authuri,user=user,req=req)
                 return user, pwd
 
         if auth_user:
@@ -144,26 +162,27 @@ class HTTPPasswordHandler(object):
             user = auth_user
             if pwd:
                 self.pwd_cache[cache_key] = user, pwd
-                self._debug_reply(ui, _("Auth data set in .hg/hgrc"),
-                                  base_url, user, pwd)
-                self.last_reply = dict(realm=realm,authuri=authuri,user=user)
+                _debug_reply(ui, _("Auth data set in .hg/hgrc"),
+                             base_url, user, pwd)
+                self.last_reply = dict(realm=realm,authuri=authuri,user=user,req=req)
                 return user, pwd
             else:
-                ui.debug(_("Username found in .hg/hgrc: %s\n" % user))
+                _debug(ui, _("Username found in .hg/hgrc: %s") % user)
 
         # Loading password from keyring.
         # Only if username is known (so we know the key) and we are
         # not after failure (so we don't reuse the bad password).
         if user and not after_bad_auth:
+            _debug(ui, _("Looking for password for user %s and url %s") % (user, keyring_url))
             pwd = password_store.get_http_password(keyring_url, user)
             if pwd:
                 self.pwd_cache[cache_key] = user, pwd
-                self._debug_reply(ui, _("Keyring password found"),
-                                  base_url, user, pwd)
-                self.last_reply = dict(realm=realm,authuri=authuri,user=user)
+                _debug_reply(ui, _("Keyring password found"),
+                             base_url, user, pwd)
+                self.last_reply = dict(realm=realm,authuri=authuri,user=user,req=req)
                 return user, pwd
             else:
-                ui.debug(_("Password not present in the keyring\n"))
+                _debug(ui, _("Password not present in the keyring"))
 
         # Is the username permanently set?
         fixed_user = (user and True or False)
@@ -188,15 +207,15 @@ class HTTPPasswordHandler(object):
             # It is done only if username is permanently set.
             # Otherwise we won't be able to find the password so it
             # does not make much sense to preserve it
-            ui.debug("Saving password for %s to keyring\n" % user)
+            _debug(ui, _("Saving password for %s to keyring") % user)
             password_store.set_http_password(keyring_url, user, pwd)
 
         # Saving password to the memory cache
         self.pwd_cache[cache_key] = user, pwd
 
-        self._debug_reply(ui, _("Manually entered password"),
-                          base_url, user, pwd)
-        self.last_reply = dict(realm=realm,authuri=authuri,user=user)
+        _debug_reply(ui, _("Manually entered password"),
+                     base_url, user, pwd)
+        self.last_reply = dict(realm=realm,authuri=authuri,user=user,req=req)
         return user, pwd
 
     def load_hgrc_auth(self, ui, base_url, user):
@@ -284,10 +303,6 @@ class HTTPPasswordHandler(object):
         return "%s://%s%s" % (parsed_url.scheme, parsed_url.netloc,
                               parsed_url.path)
 
-    def _debug_reply(self, ui, msg, url, user, pwd):
-        ui.debug("%s. Url: %s, user: %s, passwd: %s\n" % (
-            msg, url, user, pwd and '*' * len(pwd) or 'not set'))
-
 ############################################################
 
 @monkeypatch_method(passwordmgr)
@@ -303,7 +318,28 @@ def find_user_password(self, realm, authuri):
     if not hasattr(self, '_pwd_handler'):
         self._pwd_handler = HTTPPasswordHandler()
 
-    return self._pwd_handler.find_auth(self, realm, authuri)
+    if hasattr(self, '_http_req'):
+        req = self._http_req
+    else:
+        req = None
+
+    return self._pwd_handler.find_auth(self, realm, authuri, req)
+
+@monkeypatch_method(AbstractBasicAuthHandler, "http_error_auth_reqed")
+def basic_http_error_auth_reqed(self, authreq, host, req, headers):
+    self.passwd._http_req = req
+    try:
+        return basic_http_error_auth_reqed.orig(self, authreq, host, req, headers)
+    finally:
+        self.passwd._http_req = None
+
+@monkeypatch_method(AbstractDigestAuthHandler, "http_error_auth_reqed")
+def digest_http_error_auth_reqed(self, authreq, host, req, headers):
+    self.passwd._http_req = req
+    try:
+        return digest_http_error_auth_reqed.orig(self, authreq, host, req, headers)
+    finally:
+        self.passwd._http_req = None
 
 ############################################################
 
